def strtobool(value):
    v = value.lower()
    if v in ("y", "yes", "t", "true", "on", "1"):
        return True
    if v in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError



import io

from .riff import RiffChunk, RiffList
from .aep_riff import AepParser


def aepx_to_chunk(element, parser):
    header = element.tag.rsplit("}", 1)[-1].ljust(4)
    if header == "ProjectXMPMetadata":
        chunk = RiffChunk(header, 0, element.text)
    elif header == "string":
        txt = element.text or ""
        chunk = RiffChunk("Utf8", len(txt), txt)
    elif header == "numS":
        return RiffChunk(header, 0, int(element[0].text))
    elif header == "ppSn":
        return RiffChunk(header, 8, float(element[0].text))
    elif "bdata" in element.attrib:
        hex = element.attrib["bdata"]
        raw = bytes(int(hex[i:i+2], 16) for i in range(0, len(hex), 2))

        if header in parser.chunk_parsers:
            bdata = io.BytesIO(raw)
            parser.file = bdata
            data = parser.chunk_parsers[header](parser, len(raw))
        else:
            data = raw

        chunk = RiffChunk(header, len(raw), data)
    else:
        if header == "AfterEffectsProject":
            header = "RIFX"
            type = ""
        elif header in AepParser.utf8_containers:
            type = ""
        else:
            type = header
            header = "LIST"

        if header == "LIST":
            parser.on_list_start(type)

        data = RiffList(type, tuple(aepx_to_chunk(child, parser) for child in element))
        chunk = RiffChunk(header, 0, data)

        if header == "LIST":
            parser.on_list_end(type)

    parser.on_chunk(chunk)

    return chunk



from . import base, core, svg
from .base import importers

__all__ = [
    "base", "core", "svg",
    "importers",
]

try:
    from . import raster
    __all__ += ["raster"]
except ImportError:
    pass


try:
    from . import sif
    __all__ += ["sif"]
except ImportError:
    pass



import io

from xml.etree import ElementTree

from .base import importer
from ..parsers.baseporter import ExtraOption
from ..parsers.aep.aep_riff import AepParser
from ..parsers.aep.converter import AepConverter, ExpressionMode, can_convert_expressions
from ..parsers.aep.aepx import aepx_to_chunk

opts = [
    ExtraOption("comp", help="Name of the composition to extract", default=None)
]

if can_convert_expressions:
    opts.append(ExtraOption(
        "expressions", help="Export expressions", action="store_true", default=False
    ))


def convert(rifx, comp, expressions):
    conv = AepConverter(ExpressionMode.Bodymovin if expressions else ExpressionMode.Ignore)
    return conv.import_aep(rifx, comp)


@importer("AfterEffect Project", ["aep"], opts, slug="aep")
def import_aep(file, comp=None, expressions=False):
    if isinstance(file, str):
        with open(file, "rb") as fileobj:
            return import_aep(fileobj, comp, expressions)

    parser = AepParser(file)
    return convert(parser.parse(), comp, expressions)


@importer("AfterEffect Project XML", ["aepx"], opts, slug="aepx")
def import_aepx(file, comp=None, expressions=False):
    dom = ElementTree.parse(file)
    parser = AepParser(None)

    rifx = aepx_to_chunk(dom.getroot(), parser)
    return convert(rifx, comp, expressions)



import os
import enum

from ... import objects
from ...nvector import NVector
from ...utils.color import Color
from .gradient_xml import parse_gradient_xml

try:
    from .expressions import process_expression
    can_convert_expressions = True
except ImportError:
    can_convert_expressions = False
    pass


def convert_value_color(arr):
    return Color(arr[1] / 255, arr[2] / 255, arr[3] / 255, arr[0] / 255)


def shape_with_defaults(cls, **defaults):
    def callback():
        obj = cls()
        for k, v in defaults.items():
            prop = getattr(obj, k)
            if isinstance(prop, objects.properties.AnimatableMixin):
                prop.value = v
            else:
                setattr(obj, k, v)
        return obj

    return callback


class PropertyPolicyMultidim:
    def __init__(self, converter=lambda x: x):
        self.converter = converter

    def static(self, cdat):
        if len(cdat.data.value) == 1:
            return self.converter(NVector(*cdat.data.value))[0]
        else:
            return self.converter(NVector(*cdat.data.value))

    def keyframe(self, keyframe, index):
        return self.converter(NVector(*keyframe.value))


class PropertyPolicyPrepared:
    def __init__(self, values):
        self.values = values

    def static(self, cdat):
        return self.values[0]

    def keyframe(self, keyframe, index):
        return self.values[index]


class Asset:
    def __init__(self, id, name, chunk):
        self.id = id
        self.name = name
        self.chunk = chunk

    def load(self, converter):
        pass

    @property
    def lottie_id(self):
        return "asset_%s" % self.id


class FileAsset(Asset):
    def __init__(self, id, name, chunk, pin):
        super().__init__(id, name, chunk)
        self.filename = pin.find("Als2").data.find("alas").data["fullpath"]
        if not self.name:
            self.name = os.path.basename(self.filename)
        sspc = pin.find("sspc").data
        self.width = sspc.width
        self.height = sspc.height

    def to_lottie(self):
        # TODO use plain FileAsset if not an image
        asset = objects.assets.Image("asset_%s" % self.id)
        asset.width = self.width
        asset.height = self.height
        asset.path = os.path.dirname(self.filename)
        asset.file_name = os.path.basename(self.filename)
        return asset


class SolidAsset(Asset):
    def __init__(self, id, name, chunk, cdta, opti):
        super().__init__(id, name or opti.name, chunk)
        self.color = Color(opti.r, opti.g, opti.b, opti.a)


class Comp(Asset):
    def __init__(self, id, name, chunk, cdta):
        super().__init__(id, name, chunk)
        self.cdta = cdta
        self.name = ""
        self.layers = None
        self.used_assets = set()
        self.markers = []

    @property
    def lottie_id(self):
        return "precomp_%s" % self.id

    @property
    def width(self):
        return self.cdta.width

    @property
    def height(self):
        return self.cdta.height

    @property
    def time_scale(self):
        return self.cdta.time_scale

    @property
    def frame_rate(self):
        return self.cdta.frame_rate

    def to_main(self):
        anim = objects.animation.Animation()
        anim.name = self.name
        anim.width = self.cdta.width
        anim.height = self.cdta.height
        anim.frame_rate = self.cdta.frame_rate
        anim.in_point = self.cdta.start_time * self.cdta.frame_rate
        if self.cdta.end_time_num == -1:
            anim.out_point = self.cdta.comp_duration * self.cdta.frame_rate
        else:
            anim.out_point = self.cdta.end_time * self.cdta.frame_rate
        anim.layers = self.layers
        anim.markers = self.markers

        for asset in self.used_assets:
            anim.assets = [a.to_lottie() for a in self.used_assets]

        return anim

    def to_precomp(self):
        anim = objects.assets.Precomp()
        anim.name = self.name
        anim.id = self.lottie_id
        anim.frame_rate = self.cdta.frame_rate
        anim.layers = self.layers
        return anim

    def load(self, converter):
        if self.layers is None:
            self.layers = []
            converter.load_comp(self)


class ExpressionMode(enum.Enum):
    Ignore = enum.auto()
    AsIs = enum.auto()
    Bodymovin = enum.auto()


class AepConverter:
    placeholder = "-_0_/-"
    shapes = {
        "ADBE Vector Group": objects.shapes.Group,

        "ADBE Vector Shape - Group": objects.shapes.Path,
        "ADBE Vector Shape - Rect": objects.shapes.Rect,
        "ADBE Vector Shape - Star": objects.shapes.Star,
        "ADBE Vector Shape - Ellipse": objects.shapes.Ellipse,

        "ADBE Vector Graphic - Stroke": shape_with_defaults(
            objects.shapes.Stroke,
            width=2,
            color=Color(1, 1, 1),
            line_cap=objects.shapes.LineCap.Butt,
            line_join=objects.shapes.LineJoin.Miter,
            miter_limit=4,
        ),
        "ADBE Vector Graphic - Fill": shape_with_defaults(
            objects.shapes.Fill,
            color=Color(1, 0, 0),
            fill_rule=objects.shapes.FillRule.NonZero,
        ),
        "ADBE Vector Graphic - G-Fill": shape_with_defaults(
            objects.shapes.GradientFill,
            end_point=NVector(100, 0),
            fill_rule=objects.shapes.FillRule.NonZero,
        ),
        "ADBE Vector Graphic - G-Stroke": shape_with_defaults(
            objects.shapes.GradientStroke,
            end_point=NVector(100, 0),
        ),

        "ADBE Vector Filter - Merge": objects.shapes.Merge,
        "ADBE Vector Filter - Offset": objects.shapes.OffsetPath,
        "ADBE Vector Filter - PB": objects.shapes.PuckerBloat,
        "ADBE Vector Filter - Repeater": objects.shapes.Repeater,
        "ADBE Vector Filter - RC": objects.shapes.RoundedCorners,
        "ADBE Vector Filter - Trim": objects.shapes.Trim,
        "ADBE Vector Filter - Twist": objects.shapes.Twist,
        "ADBE Vector Filter - Zigzag": objects.shapes.ZigZag,
    }
    properties = {
        "ADBE Time Remapping": ("time_remapping", None),
        "ADBE Camera Aperture": ("perspective", None),

        "ADBE Vector Shape": ("shape", None),
        "ADBE Vector Shape Direction": ("direction", objects.shapes.ShapeDirection),
        "ADBE Vector Rect Roundness": ("rounded", None),
        "ADBE Vector Rect Size": ("size", None),
        "ADBE Vector Rect Position": ("position", None),
        "ADBE Vector Ellipse Size": ("size", None),
        "ADBE Vector Ellipse Position": ("position", None),

        "ADBE Vector Star Type": ("star_type", objects.shapes.StarType),
        "ADBE Vector Star Points": ("points", None),
        "ADBE Vector Star Position": ("position", None),
        "ADBE Vector Star Inner Radius": ("inner_radius", None),
        "ADBE Vector Star Outer Radius": ("outer_radius", None),
        "ADBE Vector Star Inner Roundess": ("inner_roundness", None),
        "ADBE Vector Star Outer Roundess": ("outer_roundness", None),
        "ADBE Vector Star Rotation": ("rotation", None),

        "ADBE Vector Fill Color": ("color", convert_value_color),

        "ADBE Vector Stroke Color": ("color", convert_value_color),
        "ADBE Vector Stroke Width": ("width", None),
        "ADBE Vector Stroke Miter Limit": ("animated_miter_limit", None),
        "ADBE Vector Stroke Line Cap": ("line_cap", objects.shapes.LineCap),
        "ADBE Vector Stroke Line Join": ("line_join", objects.shapes.LineJoin),

        "ADBE Vector Grad Start Pt": ("start_point", None),
        "ADBE Vector Grad End Pt": ("end_point", None),
        "ADBE Vector Grad Colors": ("colors", None),
        "ADBE Vector Grad Type": ("gradient_type", objects.shapes.GradientType),

        "ADBE Vector Merge Type": ("merge_mode", objects.shapes.MergeMode),

        "ADBE Vector Offset Amount": ("amount", None),
        "ADBE Vector Offset Line Join": ("line_join", objects.shapes.LineJoin),
        "ADBE Vector Offset Miter Limit": ("miter_limit", None),

        "ADBE Vector PuckerBloat Amount": ("amount", None),

        "ADBE Vector Repeater Copies": ("copies", None),
        "ADBE Vector Repeater Offset": ("offset", None),
        "ADBE Vector Repeater Order": ("composite", objects.shapes.Composite),
        "ADBE Vector Repeater Anchor Point": ("anchor_point", None),
        "ADBE Vector Repeater Position": ("position", None),
        "ADBE Vector Repeater Rotation": ("rotation", None),
        "ADBE Vector Repeater Start Opacity": ("start_opacity", lambda v: v * 100),
        "ADBE Vector Repeater End Opacity": ("end_opacity", lambda v: v * 100),
        "ADBE Vector Repeater Scale": ("scale", lambda v: v * 100),

        "ADBE Vector RoundCorner Radius": ("radius", None),

        "ADBE Vector Trim Start": ("start", None),
        "ADBE Vector Trim End": ("end", None),
        "ADBE Vector Trim Offset": ("offset", None),

        "ADBE Vector Twist Angle": ("angle", None),
        "ADBE Vector Twist Center": ("center", None),

        "ADBE Vector Zigzag Size": ("amplitude", None),
        "ADBE Vector Zigzag Detail": ("frequency", None),

        "ADBE Anchor Point": ("anchor_point", None),
        "ADBE Position": ("position", None),
        "ADBE Rotate Z": ("rotation", None),
        "ADBE Opacity": ("opacity", lambda v: v * 100),
        "ADBE Scale": ("scale", lambda v: v * 100),

        "ADBE Vector Anchor Point": ("anchor_point", None),
        "ADBE Vector Position": ("position", None),
        "ADBE Vector Rotation": ("rotation", None),
        "ADBE Vector Group Opacity": ("opacity", None),
        "ADBE Vector Scale": ("scale", None),
    }
    property_groups = {
        "ADBE Camera Options Group": ["", None],
        "ADBE Text Properties": ["data", None],
        "ADBE Effect Parade": ["effects", "load_effects"],
        "ADBE Transform Group": ["transform", None],
        "ADBE Vector Transform Group": ["transform", None],
        "ADBE Vector Stroke Dashes": ["", "load_dashes"],
        "ADBE Text Path Options": ["masked_path", None],
        "ADBE Text More Options": ["more_options", None],
        "ADBE Vector Repeater Transform": ["transform", None],
    }
    effect_value_types = {
        0: objects.effects.EffectValueLayer,
        2: objects.effects.EffectValueSlider,
        3: objects.effects.EffectValueAngle,
        4: objects.effects.EffectValueCheckbox,
        5: objects.effects.EffectValueColor,
        6: objects.effects.EffectValuePoint,
        7: objects.effects.EffectValueDropDown,
        9: objects.effects.EffectNoValue,
        10: objects.effects.EffectValueSlider,
        13: objects.effects.CustomEffect,
        15: objects.effects.EffectNoValue,
        16: objects.effects.EffectValuePoint,
    }
    effect_match_names = {
        "ADBE Tint": objects.effects.TintEffect,
        "ADBE Fill": objects.effects.FillEffect,
        "ADBE Stroke": objects.effects.StrokeEffect,
        "ADBE Tritone": objects.effects.TritoneEffect,
        "ADBE Pro Levels2": objects.effects.ProLevelsEffect,
        "ADBE Drop Shadow": objects.effects.DropShadowEffect,
        "ADBE Radial Wipe": objects.effects.RadialWipeEffect,
        "ADBE Displacement Map": objects.effects.DisplacementMapEffect,
        "ADBE Set Matte3": objects.effects.Matte3Effect,
        "ADBE Gaussian Blur 2": objects.effects.GaussianBlurEffect,
        "ADBE Twirl": objects.effects.TwirlEffect,
        "ADBE MESH WARP": objects.effects.MeshWarpEffect,
        "ADBE Ripple": objects.effects.WavyEffect,
        "ADBE Spherize": objects.effects.SpherizeEffect,
        "ADBE FreePin3": objects.effects.PuppetEffect,
    }

    def __init__(self, expression_mode=ExpressionMode.Ignore):
        self.time_mult = 1
        self.time_offset = 0
        self.assets = {}
        self.comps = {}
        self.layers = {}
        self.effects = {}
        self.expression_mode = expression_mode

    def read_properties(self, object, chunk):
        match_name = None
        for item in chunk.data.children:
            # Match name
            if item.header == "tdmn":
                match_name = item.data
            elif match_name in self.property_groups:
                prop, funcn = self.property_groups[match_name]
                subobj = object if not prop else getattr(object, prop)
                func = self.read_properties if funcn is None else getattr(self, funcn)
                func(subobj, item)
            # Name
            elif item.header == "tdsn" and len(item.data.children) > 0:
                name = item.data.children[0]
                if name.header == "Utf8" and name.data != self.placeholder and name.data:
                    object.name = name.data
            # Shape hidden
            elif item.header == "tdsb":
                if (item.data & 1) == 0:
                    object.hidden = True
            # MultiDimensional property
            elif item.header == "LIST" and item.data.type == "tdbs":
                self.parse_property_multidimensional(object, match_name, item)
            # Shape property
            elif item.header == "LIST" and item.data.type == "om-s":
                self.parse_property_shape(object, match_name, item)
            # Gradient color property
            elif item.header == "LIST" and item.data.type == "GCst":
                self.parse_property_gradient(object, match_name, item)
            # Sub-object
            elif item.header == "LIST" and item.data.type == "tdgp":
                if match_name == "ADBE Vectors Group" or match_name == "ADBE Root Vectors Group":
                    self.read_properties(object, item)
                elif match_name == "ADBE Vector Transform Group" or match_name == "ADBE Transform Group":
                    self.read_properties(object.transform, item)
                elif match_name in self.shapes:
                    child = self.shapes[match_name]()
                    object.add_shape(child)
                    child.match_name = match_name
                    self.read_properties(child, item)

    def parse_property_multidimensional(self, object, match_name, chunk):
        meta = self.properties.get(match_name)
        if not meta:
            return

        prop_name, converter = meta
        policy = PropertyPolicyMultidim()
        if converter is not None:
            policy.converter = converter

        prop = objects.properties.MultiDimensional()
        self.parse_property_tdbs(chunk, prop, policy)

        setattr(object, prop_name, prop)

    def parse_property_tdbs(self, chunk, prop, policy):
        tdb4, static, kf, expr = chunk.data.find_multiple("tdb4", "cdat", "list", "Utf8")

        if static:
            prop.value = policy.static(static)

        if kf:
            self.set_property_keyframes(prop, policy, kf, tdb4)

        if expr:
            if self.expression_mode == ExpressionMode.AsIs:
                prop.expression = expr.data
            elif self.expression_mode == ExpressionMode.Bodymovin:
                prop.expression = process_expression(expr.data)

    def time(self, value):
        return (value + self.time_offset) * self.time_mult

    def set_property_keyframes(self, prop, policy, chunk, tdb4):
        ldat = chunk.data.find("ldat")
        if ldat and hasattr(ldat.data, "keyframes"):
            if len(ldat.data.keyframes) == 0:
                return

            for index, aep_kf in enumerate(ldat.data.keyframes):
                time = self.time(aep_kf.time_num / tdb4.time_den)

                kf = prop.add_keyframe(
                    time,
                    policy.keyframe(aep_kf, index)
                )

                if aep_kf.hold:
                    kf.hold = True
                elif not aep_kf.ease:
                    objects.easing.Linear()(kf)
                else:
                    self.keyframe_ease(policy, aep_kf, kf, ldat.data.keyframes, index, tdb4)

        if len(prop.keyframes) == 1:
            prop.clear_animation(prop.keyframes[0].start)

    def _keyframe_pos_tan(self, length, tan):
        if len(tan) > length:
            tan = tan[:length]
        else:
            tan = tan + ([0] * length - len(tan))
        return NVector(tan)

    def keyframe_ease(self, policy, aep_kf, kf, keyframes, index, tdb4):
        next_i = (index + 1) % len(keyframes)
        next_aep_kf = keyframes[next_i]
        next_value = policy.keyframe(next_aep_kf, next_i)

        if tdb4.position:
            kf.out_tan = self._keyframe_pos_tan(len(kf.value), aep_kf.pos_tan_out)
            kf.in_tan = self._keyframe_pos_tan(len(kf.value), next_aep_kf.pos_tan_in)

        next_time = self.time(next_aep_kf.time)
        duration = next_time - kf.time
        if duration == 0:
            return objects.easing.Linear()(kf)

        in_infl = next_aep_kf.in_influence
        in_speed = next_aep_kf.in_speed
        out_infl = aep_kf.out_influence
        out_speed = aep_kf.out_speed

        if tdb4.position or tdb4.no_value:
            if isinstance(in_infl, list):
                in_infl = in_infl[0]
                in_speed = in_speed[0]
                out_infl = out_infl[0]
                out_speed = out_speed[0]

            if tdb4.no_value:
                curve_length = 1
            else:
                curve_length = objects.bezier.CubicBezierSegment(
                    kf.value, kf.out_tan, kf.in_tan, next_value
                ).length
            average_speed = curve_length / duration
            if curve_length == 0:
                out_bez = out_infl / 100
                in_bez = in_infl / 100
            else:
                out_bez = min(curve_length / (out_speed * duration), out_infl / 100)
                in_bez = min(curve_length / (in_speed * duration), in_infl / 100)

            kf.in_value.x = 1 - in_bez
            kf.out_value.x = out_bez

            # TODO fuzzy compare?
            if average_speed == 0:
                kf.in_value.y = kf.in_value.x
                kf.out_value.y = kf.out_value.x
            else:
                kf.in_value.y = 1 - in_speed / average_speed * in_bez
                kf.out_value.y = kf.out_value.x / average_speed * out_bez

        else:
            in_x = []
            out_x = []
            in_y = []
            out_y = []

            for i in range(tdb4.components):
                in_x.append(1 - in_infl[i] / 100)
                out_x.append(out_infl[i] / 100)

                y_normal = next_value[i] - kf.value[i]
                if tdb4.color:
                    y_normal *= 255

                if abs(y_normal) < 0.0000001:
                    y_normal = 1

                out_bez_y = out_speed[i] * out_infl[i] / 100
                in_bez_y = in_speed[i] * in_infl[i] / 100
                out_y.append(out_bez_y * duration / y_normal)
                in_y.append(1 - in_bez_y * duration / y_normal)

            kf.in_value.x = in_x
            kf.in_value.y = in_y
            kf.out_value.x = out_x
            kf.out_value.y = out_y

    def parse_property_shape(self, object, match_name, chunk):
        meta = self.properties.get(match_name)
        if not meta:
            return

        prop_name = meta[0]

        prop = objects.properties.ShapeProperty()

        policy = PropertyPolicyPrepared([])
        tdbs, omks = chunk.data.find_multiple("tdbs", "omks")

        self.parse_shape_omks(omks, policy)
        self.parse_property_tdbs(tdbs, prop, policy)

        setattr(object, prop_name, prop)

    def parse_shape_omks(self, chunk, policy):
        for item in chunk.data.children:
            if item.header == "LIST" and item.data.type == "shap":
                policy.values.append(self.parse_shape_shap(item))

    def parse_property_gradient(self, object, match_name, chunk):
        meta = self.properties.get(match_name)
        if not meta:
            return

        colors = getattr(object, meta[0])
        prop = colors.colors

        policy = PropertyPolicyPrepared([])
        tdbs, keys = chunk.data.find_multiple("tdbs", "GCky")

        for item in keys.data.children:
            if item.header == "Utf8":
                policy.values.append(parse_gradient_xml(item.data, colors))

        self.parse_property_tdbs(tdbs, prop, policy)

    def parse_shape_shap(self, chunk):
        bez = objects.bezier.Bezier()
        shph, list = chunk.data.find_multiple("shph", "list")

        top_left = shph.data.top_left
        bottom_right = shph.data.bottom_right
        bez.closed = not shph.data.open

        points = list.data.find("ldat").data.points
        for i in range(0, len(points), 3):
            vertex = self.absolute_bezier_point(top_left, bottom_right, points[i])
            tan_in = self.absolute_bezier_point(top_left, bottom_right, points[(i-1) % len(points)])
            tan_ou = self.absolute_bezier_point(top_left, bottom_right, points[i+1])
            print(i, vertex, tan_in, tan_ou)
            bez.add_point(vertex, tan_in - vertex, tan_ou - vertex)
        return bez

    def absolute_bezier_point(self, tl, br, p):
        return NVector(
            tl[0] * (1-p.x) + br[0] * p.x,
            tl[1] * (1-p.y) + br[1] * p.y
        )

    def load_dashes(self, stroke, chunk):
        stroke.dashes = []
        mn = None
        for item in chunk.data.children:
            if item.header == "tdmn":
                mn = item.data
            elif item.header == "LIST" and item.data.type == "tdbs":
                dash = objects.shapes.StrokeDash()
                self.parse_property_tdbs(item, dash.length, PropertyPolicyMultidim())
                dash.match_name = mn
                name = mn[len("ADBE Vector Stroke "):].replace(" ", "").lower()
                dash.name = name
                dash.type = objects.shapes.StrokeDashType(name[0])
                stroke.dashes.append(dash)

    def load_effect_values(self, effect, chunk):
        for index, tdbs in enumerate(chunk.data.find_all("tdbs")):
            if index == 0:
                continue
            policy = PropertyPolicyMultidim()
            prop = effect.effects[index-1].value
            self.parse_property_tdbs(tdbs, prop, policy)

    def load_effects(self, layer: objects.layers.VisualLayer, chunk):
        mn = None
        layer.effects = []

        for item in chunk.data.children:
            if item.header == "tdmn":
                mn = item.data
            elif item.header == "LIST" and item.data.type == "sspc":
                effect = self.effects[mn].clone()
                self.load_effect_values(effect, item.data.find("tdgp"))
                layer.effects.append(effect)

    def create_asset_layer(self, ldta):
        asset = self.assets[ldta.source_id]
        if isinstance(asset, Comp):
            layer = objects.layers.PreCompLayer()
            layer.reference_id = asset.lottie_id
            return layer
        elif isinstance(asset, SolidAsset):
            layer = objects.layers.SolidColorLayer()
            layer.color = asset.color
        elif asset.width > 0 and asset.height > 0:
            layer = objects.layers.ImageLayer()
            layer.reference_id = asset.lottie_id
        else:
            layer = objects.layers.AudioLayer()
            layer.reference_id = asset.lottie_id

        layer.width = asset.width
        layer.height = asset.height
        layer.name = asset.name
        return layer

    def chunk_to_layer(self, chunk):
        name = None
        layer = None
        for item in chunk.data.children:
            if item.header == "Utf8":
                if item.data:
                    layer.name = item.data
            elif item.header == "ldta":
                # Asset layer
                if item.data.type == 0:
                    if item.data.null:
                        layer = objects.layers.NullLayer()
                    else:
                        layer = self.create_asset_layer(item)
                # Light Layer
                elif item.data.type == 1:
                    return None
                # Camera Layer
                elif item.data.type == 2:
                    layer = objects.layers.CameraLayer()
                elif item.data.type == 3:
                    layer = objects.layers.TextLayer()
                elif item.data.type == 4:
                    layer = objects.layers.ShapeLayer()

                layer.name = name
                if isinstance(layer, objects.layers.VisualLayer):
                    layer.transform.position.value = NVector(self.anim.width / 2, self.anim.height / 2)
                    if item.data.parent_id:
                        layer.parent_index = item.data.parent_id
                self.time_offset = item.data.start_time
                layer.start_time = self.time_offset * self.time_mult
                layer.in_point = self.time(item.data.in_time)
                layer.out_point = self.time(item.data.out_time)
                layer.threedimensional = item.data.ddd
                layer.hidden = not item.data.visible
                layer.index = item.data.layer_id
                layer.is_guide = item.data.guide
                layer.matte_mode = objects.layers.MatteMode(item.data.matte_mode)

            elif item.header == "LIST":
                self.read_properties(layer, item)

        return layer

    def chunk_to_markerts(self, chunk, comp: Comp):
        mrst = chunk.data.find("tdgp").data.find("mrst").data

        mrky, tdbs = mrst.find_multiple("mrky", "tdbs")

        markers = []

        for item in mrky.data.children:
            if item.header == "LIST" and item.data.type == "Nmrd":
                marker = objects.animation.Marker()
                nmhd, comment = item.data.find_multiple("NmHd", "Utf8")
                marker.duration = nmhd.data.duration * comp.frame_rate
                marker.comment = comment.data
                markers.append(marker)

        for i, kf in enumerate(tdbs.data.find("list").data.find("ldat").data.keyframes):
            markers[i].time = self.time(kf.time)

        comp.markers += markers

    def load_comp(self, comp: Comp):
        self.anim = comp
        self.time_mult = comp.frame_rate
        self.time_offset = 0

        for item in comp.chunk.data.children:
            if item.header == "LIST":
                if item.data.type == "Layr":
                    layer = self.chunk_to_layer(item)
                    if layer:
                        comp.layers.append(layer)
                elif item.data.type == "SecL":
                    self.chunk_to_markerts(item, comp)

        return comp

    def collect_assets(self, fold):
        for chunk in fold.data.children:
            if chunk.header == "LIST" and chunk.data.type == "Item":
                item_data = chunk.data.find("idta").data
                name = chunk.data.find("Utf8").data
                if item_data.type == 1:
                    self.items_from_fold(chunk)
                    return
                elif item_data.type == 4:
                    asset = Comp(item_data.id, name, chunk, chunk.data.find("cdta").data)
                    self.comps[name] = asset
                elif item_data.type == 7:
                    pin = chunk.find("Pin ").data
                    opti = pin.find("opti")

                    if opti.type == "Soli":
                        asset = SolidAsset(item_data.id, name, chunk, opti)
                    else:
                        asset = FileAsset(item_data.id, name, chunk, pin)

                self.assets[asset.id] = asset

    def collect_effects(self, effects):
        for definition in effects.data.find_all("EfDf"):
            match, params = definition.data.find_multiple("tdmn", "sspc")
            if match in self.effects:
                continue

            if match in self.effect_match_names:
                effect = self.effect_match_names()
                effect.name = params.data.find("fnam").data.find("Utf8").data
            else:
                effect = objects.effects.CustomEffect()
                index = 0
                mn = None
                for chunk in params.data.children:
                    if chunk.header == "tdmn":
                        mn = chunk.data
                    elif chunk.header == "fnam":
                        effect.name = chunk.data.find("Utf8").data
                    elif chunk.header == "LIST" and chunk.data.type == "parT":
                        index += 1
                        if index == 1:
                            continue
                        pard = chunk.data.find("pard").data
                        type = pard.value
                        val = self.effect_value_types.get(type, objects.effects.EffectNoValue)()
                        val.match_name = mn
                        val.name = pard.name
                        effect.effects.append(val)

            effect.match_name = match
            self.effects[match] = effect

    def process(self, top_level):
        fold, effects = top_level.data.find_multiple("Fold", "EfdG")
        self.collect_assets(fold)
        if effects:
            self.collect_effects(effects)

    def import_aep(self, top_level, name):
        self.process(top_level)

        if not name:
            comp = next(iter(self.comps.values()))
        else:
            comp = self.comps[name]

        comp.load(self)

        return comp.to_main()



import struct
from dataclasses import dataclass


class Endianness:
    def read(self, file, size):
        return self.decode(file.read(size))

    def decode(self, data):
        raise NotImplementedError()

    def decode_2comp(self, data):
        uint = self.decode(data)
        sbit = 1 << (len(data) * 8 - 1)
        if uint & sbit:
            return uint - (sbit << 1)
        else:
            return uint


class BigEndian(Endianness):
    @staticmethod
    def decode_data(data):
        value = 0
        for byte in data:
            value <<= 8
            value |= byte

        return value

    def decode(self, data):
        return self.decode_data(data)

    def decode_float64(self, data):
        return struct.unpack(">d", data)[0]

    def decode_float32(self, data):
        return struct.unpack(">f", data)[0]


class LittleEndian(Endianness):
    def decode(self, data):
        return BigEndian.decode_data(reversed(data))

    def decode_float64(self, data):
        return struct.unpack("<d", data)[0]

    def decode_float32(self, data):
        return struct.unpack("<f", data)[0]


@dataclass
class RiffChunk:
    header: str
    length: int
    data: bytes


@dataclass
class RiffList:
    type: str
    children: tuple

    def find(self, header):
        for ch in self.children:
            if ch.header == header:
                return ch
            elif ch.header == "LIST" and ch.data.type == header:
                return ch

    def find_list(self, type):
        for ch in self.children:
            if ch.header == "LIST" and ch.data.type == type:
                return ch

    def find_multiple(self, *headers):
        headers = list(headers)
        found = [None] * len(headers)
        for ch in self.children:
            for i, header in enumerate(headers):
                if ch.header == header:
                    headers[i] = None
                    found[i] = ch
                    break
                elif ch.header == "LIST" and ch.data.type == header:
                    headers[i] = None
                    found[i] = ch
                    break

        return found


sint = object()


@dataclass
class RiffHeader:
    endianness: Endianness
    length: int
    format: str


class RiffParser:
    def __init__(self, file):
        self.file = file
        magic = self.read(4)
        if magic == b"RIFF":
            endian = LittleEndian()
        elif magic == b"RIFX":
            endian = BigEndian()
        else:
            raise Exception("Expected RIFF or RIFX")

        self.endian = endian
        length = self.read_number(4)
        self.end = file.tell() + length
        format = self.read_str(4)

        self.header = RiffHeader(self.endian, length, format)
        self.chunk_parsers = {}
        self.weird_lists = {}

    def read(self, length):
        return self.file.read(length)

    def read_str(self, length):
        return self.read(length).decode("ascii")

    def read_number(self, length):
        return self.endian.read(self.file, length)

    def read_float(self, length):
        if length == 4:
            return self.endian.decode_float32(self.read(4))
        elif length == 8:
            return self.endian.decode_float64(self.read(8))
        else:
            raise TypeError("Invalid float size %s" % length)

    def read_chunk(self, chunk_max_end):
        header = self.read_str(4)

        length = self.read_number(4)

        if length + self.file.tell() > chunk_max_end:
            length = chunk_max_end - self.file.tell()

        if header == "LIST":
            end = self.file.tell() + length
            type = self.read_str(4)
            if type in self.weird_lists:
                data = StructuredData()
                data.type = type
                data.data = self.weird_lists[type](self, length-4)
            else:
                self.on_list_start(type)
                children = []
                while self.file.tell() < end:
                    children.append(self.read_chunk(end))
                data = RiffList(type, tuple(children))
                self.on_list_end(type)
        elif header in self.chunk_parsers:
            data = self.chunk_parsers[header](self, length)
        else:
            data = self.read(length)

        if data is None:
            raise Exception("Incomplete chunk")

        chunk = RiffChunk(header, length, data)

        self.on_chunk(chunk)

        # Skip pad byte
        if length % 2:
            self.read(1)

        return chunk

    def __iter__(self):
        while True:
            if self.file.tell() >= self.end:
                return

            yield self.read_chunk(self.end)

    def on_chunk(self, chunk):
        pass

    def on_list_start(self, type):
        pass

    def on_list_end(self, type):
        pass

    def read_sub_chunks(self, length):
        end = self.file.tell() + length
        children = []
        while self.file.tell() < end:
            children.append(self.read_chunk(end))
        return RiffList("", tuple(children))

    def parse(self):
        return RiffChunk(
            "RIFX" if isinstance(self.endian, BigEndian) else "RIFF",
            self.end,
            RiffList("", tuple(self))
        )


class StructuredData:
    def __init__(self):
        self.raw_bytes = b''


class StructuredReader:
    def __init__(self, parser, length):
        self.value = StructuredData()
        self.index = 0
        self.parser = parser
        self.length = length
        self.to_read = length

    def skip(self, byte_count):
        self.read_attribute("", byte_count, bytes)

    def read_attribute_string0(self, name, length):
        self.set_attribute(name, self.read_string0(length))

    def set_attribute(self, name, value):
        if name == "":
            name = "_%s" % self.index
            self.index += 1

        setattr(self.value, name, value)

    def read_attribute(self, name, size, type):
        self.set_attribute(name, self.read_value(size, type))

    def read_attribute_array(self, name, count, length, type):
        self.set_attribute(name, self.read_array(count, length, type))

    def finalize(self):
        if self.to_read:
            self.skip(self.to_read)

    def read_raw(self, length):
        raw = self.parser.read(length)
        self.to_read -= length
        self.value.raw_bytes += raw
        return raw

    def read_string0(self, length):
        read = self.read_raw(length)

        try:
            read = read[:read.index(b'\0')]
        except ValueError:
            pass

        try:
            return read.decode("utf8")
        except UnicodeDecodeError:
            return read

    def read_array(self, count, length, type):
        value = []
        for i in range(count):
            value.append(self.read_value(length, type))
        return value

    def read_value(self, length, type):
        if isinstance(type, list):
            val = []
            for name, size, subtype in type:
                val.append(self.read_value(size, subtype))
            return val

        if length > self.to_read:
            raise Exception("Not enough data in chunk")

        data = self.read_raw(length)

        if type is bytes:
            return data
        elif type is int:
            return self.parser.endian.decode(data)
        elif type is sint:
            return self.parser.endian.decode_2comp(data)
        elif type is str:
            return data.decode("utf8")
        elif type is float:
            if length == 8:
                return self.parser.endian.decode_float64(data)
            elif length == 4:
                return self.parser.endian.decode_float32(data)
            else:
                TypeError("Wrong size for float: %s" % length)
        else:
            raise TypeError("Unknown value type %s" % type)

    def attr_bit(self, name, byte, bit, attr="attrs"):
        setattr(self.value, name, (getattr(self.value, attr)[byte] & (1 << bit)) != 0)

    def fraction(self, name, num="_num", den="_den"):
        setattr(self.value, name, getattr(self.value, name + num) / getattr(self.value, name + den))



import io
import enum
import dataclasses


class TokenType(enum.Enum):
    # /foo
    Identifier = enum.auto()
    # 123
    Number = enum.auto()
    # (foo)
    String = enum.auto()
    # <f000>
    HexString = enum.auto()
    # true
    Boolean = enum.auto()
    # <<
    ObjectStart = enum.auto()
    # >>
    ObjectEnd = enum.auto()
    # [
    ArrayStart = enum.auto()
    # ]
    ArrayEnd = enum.auto()
    # null
    Null = enum.auto()
    # end of file
    Eof = enum.auto()
    # obj
    IndirectObjectStart = enum.auto()
    # endobj
    IndirectObjectEnd = enum.auto()
    # R
    IndirectReference = enum.auto()
    # stream...endstream
    Stream = enum.auto()


class Token:
    def __init__(self, type, value=None):
        self.type = type
        self.value = value

    def __str__(self):
        return "(%s, %s)" % (self.type.name, self.value)

    def __repr__(self):
        return "<Token %s>" % self


@dataclasses.dataclass
class IndirectObject:
    object_number: int
    generation_number: int
    data: dict


@dataclasses.dataclass
class IndirectReference:
    object_number: int
    generation_number: int


@dataclasses.dataclass
class Stream:
    dictionary: dict
    data: bytes


class CosParser:
    def __init__(self, file, max_pos=None):
        self.file = file
        self.max_pos = max_pos
        self.lookahead = None

    def parse(self):
        self.lex()
        if self.lookahead.type == TokenType.Identifier:
            return self.parse_dict_content()

        val = self.parse_value()
        if self.lookahead.type == TokenType.Eof:
            return val

        return [val] + self.parse_array_content()

    def parse_value(self):
        if (
            self.lookahead.type == TokenType.String or
            self.lookahead.type == TokenType.HexString or
            self.lookahead.type == TokenType.Null or
            self.lookahead.type == TokenType.Boolean or
            self.lookahead.type == TokenType.Identifier or
            self.lookahead.type == TokenType.Stream
        ):
            val = self.lookahead.value
            self.lex()
            return val

        if self.lookahead.type == TokenType.Number:
            val = self.lookahead.value
            self.lex()
            if self.lookahead.type == TokenType.Number:
                gen = self.lookahead.value
                state = self.save_state()
                self.lex()
                if self.lookahead.type == TokenType.IndirectObjectStart:
                    self.lex()
                    data = self.parse_value()
                    self.expect(TokenType.IndirectObjectEnd)
                    self.lex()
                    return IndirectObject(val, gen, data)
                elif self.lookahead.type == TokenType.IndirectReference:
                    self.lex()
                    return IndirectReference(val, gen)
                else:
                    self.restore_state(state)
            return val

        if self.lookahead.type == TokenType.ObjectStart:
            self.lex()
            val = self.parse_dict_content()
            self.expect(TokenType.ObjectEnd)
            self.lex()
            if self.lookahead.type == TokenType.Stream:
                val = Stream(val, self.lookahead.value)
                self.lex()
            return val

        if self.lookahead.type == TokenType.ArrayStart:
            self.lex()
            val = self.parse_array_content()
            self.expect(TokenType.ArrayEnd)
            self.lex()
            return val

        raise SyntaxError("Expected COS value, got %s" % self.lookahead)

    def save_state(self):
        return (self.max_pos, self.file.tell(), self.lookahead)

    def restore_state(self, state):
        self.max_pos, pos, self.lookahead = state
        self.file.seek(pos)

    def parse_dict_content(self):
        value = {}
        while True:
            if self.lookahead.type == TokenType.Eof or self.lookahead.type == TokenType.ObjectEnd:
                break

            self.expect(TokenType.Identifier)
            key = self.lookahead.value
            self.lex()
            val = self.parse_value()
            value[key] = val

        return value

    def parse_array_content(self):
        value = []
        while True:
            if self.lookahead.type == TokenType.Eof or self.lookahead.type == TokenType.ArrayEnd:
                break
            value.append(self.parse_value())
        return value

    def parse_indirect(self, object_number, generation_number):
        self.lex()
        if self.lookahead.type == TokenType.IndirectObjectStart:
            self.lex()
            data = self.parse_value()
            self.expect(TokenType.IndirectObjectEnd)
            self.lex()
            return IndirectObject(object_number, generation_number, data)
        elif self.lookahead.type == TokenType.IndirectReference:
            self.lex()
            return IndirectReference(object_number, generation_number)
        else:
            raise SyntaxError("Expected `obj` or `R`, got %s" % self.lookahead)

    def expect(self, token_type):
        if self.lookahead.type != token_type:
            raise SyntaxError("Expected %s, got %s" % (token_type, self.lookahead))

    def lex(self):
        self.lookahead = self.lex_token()

    def lex_token(self):
        while True:
            char = self.get_char()
            if char is None:
                return Token(TokenType.Eof)
            elif char == b"%":
                self.lex_comment()
            elif not char.isspace():
                break

        # <<
        if char == b"<":
            char = self.get_char()
            if char == b"<":
                return Token(TokenType.ObjectStart)
            elif char is None:
                self.raise_lex(b"<")
            elif self.is_hex(char):
                return self.lex_hex_string(char)
            else:
                self.raise_lex(b"<" + char)

        # >>
        if char == b">":
            self.expect_char(b">", char)
            return Token(TokenType.ObjectEnd)

        # [
        if char == b"[":
            return Token(TokenType.ArrayStart)

        # ]
        if char == b"]":
            return Token(TokenType.ArrayEnd)

        # /foo
        if char == b"/":
            return self.lex_identifier()

        # (foo)
        if char == b"(":
            return self.lex_string()

        # Keyword
        if char.isalpha():
            return self.lex_keyword(char)

        # Number
        if char.isdigit() or char in b".+-":
            return self.lex_number(char)

        self.raise_lex(char)

    def expect_char(self, exp, head):
        char = self.get_char()
        if char != exp:
            self.raise_lex(head+char, head+exp)

    def raise_lex(self, token, exp=None):
        msg = "Unknown COS token %s" % token
        if exp is not None:
            msg += ", expected %s" % exp
        raise SyntaxError(msg)

    def get_char(self):
        if self.max_pos is None:
            return self.file.read(1)

        if self.max_pos <= 0:
            return None

        char = self.file.read(1)
        self.max_pos -= 1
        return char

    def unget(self):
        self.file.seek(-1, io.SEEK_CUR)
        if self.max_pos is not None:
            self.max_pos += 1

    # Skip until newline
    def lex_comment(self):
        while True:
            char = self.get_char()
            if char == b'\n':
                break
            elif char is None:
                break

    def lex_number(self, char):
        if char == b'.':
            return self.lex_num_fract(self.get_char(), char)
        elif char == b'+' or char == b'-':
            return self.lex_num_int(self.get_char(), char)
        else:
            return self.lex_num_int(char, b'')

    def lex_num_int(self, char, head):
        while True:
            if char == b'.':
                return self.lex_num_fract(self.get_char(), head + char)
            elif char is None:
                break
            elif char.isdigit():
                head += char
                char = self.get_char()
            else:
                self.unget()
                break
        return Token(TokenType.Number, int(head))

    def lex_num_fract(self, char, head):
        while True:
            if char is None:
                break
            elif char.isdigit():
                head += char
                char = self.get_char()
            else:
                self.unget()
                break
        return Token(TokenType.Number, float(head))

    def lex_keyword(self, char):
        kw = char
        while True:
            char = self.get_char()
            if char is None:
                break
            elif char.isalpha():
                kw += char
            else:
                self.unget()
                break

        if kw == b"true":
            return Token(TokenType.Boolean, True)
        elif kw == b"false":
            return Token(TokenType.Boolean, False)
        elif kw == b"null":
            return Token(TokenType.Boolean, None)
        elif kw == b"obj":
            return Token(TokenType.IndirectObjectStart)
        elif kw == b"endobj":
            return Token(TokenType.IndirectObjectEnd)
        elif kw == b"R":
            return Token(TokenType.IndirectReference)
        elif kw == b"stream":
            return self.lex_stream()
        elif kw == b"xref":
            return Token(TokenType.Eof)
        else:
            raise SyntaxError("Unknown keyword %s" % kw)

    def lex_stream(self):
        char = self.get_char()
        if char == b'\r':
            if self.get_char() != b'\n':
                raise SyntaxError("Invalid newline")
        elif char != b'\n':
            raise SyntaxError("Expected newline after `stream`")

        stream = b''
        marker = b'endstream'
        while True:
            char = self.get_char()
            if char is None:
                raise SyntaxError("Unterminated stream")
            stream += char
            if stream.endswith(marker):
                break

        return Token(TokenType.Stream, stream[:-len(marker)])

    def lex_string(self):
        string = b''
        encoding = "utf-8"

        for i in range(2):
            char = self.lex_string_char()
            if char is None:
                break
            string += char

        bom = string
        if bom == b'\xfe\xff':
            encoding = "utf-16-be"
        elif bom == b'\xff\xfe':
            encoding = "utf-16-le"

        while True:
            char = self.lex_string_char()
            if char is None:
                break
            string += char

        try:
            return Token(TokenType.String, string.decode(encoding))
        except UnicodeDecodeError:
            return Token(TokenType.String, bom + string)

    def lex_string_char(self):
        char = self.get_char()
        if char is None:
            raise SyntaxError("Unterminated string")
        elif char == b')':
            return None
        elif char == b'\\':
            return self.lex_string_escape()
        elif char == b'\r':
            if self.get_char() != b'\n':
                self.unget()
            return b'\n'
        elif char == b'\n':
            if self.get_char() != b'\r':
                self.unget()
            return b'\n'
        else:
            return char

    def lex_string_escape(self):
        char = self.get_char()
        if char is None:
            raise SyntaxError("Unterminated string")

        if char == b'n':
            return b'\n'
        elif char == b'r':
            return b'\r'
        elif char == b'b':
            return b'\b'
        elif char == b'f':
            return b'\f'
        elif char == b'(':
            return b'('
        elif char == b')':
            return b')'
        elif char == b'\\':
            return b'\\'
        elif self.is_octal(char):
            octal = char[0] - b'0'[0]
            for i in range(2):
                char = self.get_char()
                if char is None:
                    break
                elif not self.is_octal(char):
                    self.unget()
                    break
                octal = octal*8 + char - b'0'[0]
            return octal.to_bytes(1, "big")

        raise SyntaxError("Invalid escape sequence")

    def is_octal(self, char):
        return b'0' <= char <= b'8'

    def is_hex(self, char):
        return char.isdigit() or b'a' <= char <= b'f' or b'A' <= char <= b'F'

    def lex_hex_string(self, hstr):
        while True:
            char = self.get_char()
            if char is None:
                raise SyntaxError("Unterminated hex string")
            elif char.is_hex():
                hstr += char
            elif char == b'>':
                break
            elif not char.isspace():
                raise SyntaxError("Invalid character in hex string: %s" % char)

        if len(hstr) % 2:
            hstr += b'0'

        data = b''
        for i in range(0, len(hstr), 2):
            data += int(hstr[i:i+1], 16).to_bytes(1, "big")

        return Token(TokenType.HexString, data)

    def lex_identifier(self):
        ident = ""
        special = b'()[]<>{}/%'
        while True:
            char = self.get_char()
            if char is None:
                break
            elif char < b'!' or char > b'~':
                self.unget()
                break
            elif char == b'#':
                hexstr = b''
                for i in range(2):
                    char = self.get_char()
                    if char is None or not self.is_hex(char):
                        raise SyntaxError("Invalid identifier")
                    hexstr += char
                ident += chr(int(hexstr, 16))
            elif char in special:
                self.unget()
                break
            else:
                ident += chr(char[0])

        return Token(TokenType.Identifier, ident)



import esprima
import escodegen


class Converter:
    reserved_properties = ["position", "scale", "anchorPoint", "rotation"]
    op_to_func = {
        "+": "$bm_sum",
        "-": "$bm_sub",
        "*": "$bm_mul",
        "/": "$bm_div",
        "%": "$bm_mod",
    }

    def __init__(self):
        self.stack = [set()]
        self.scope = set()
        self.undeclared = set()

    def value_to_node(self, val):
        if isinstance(val, list):
            return list(map(self.value_to_node, val))
        elif isinstance(val, dict):
            node = esprima.nodes.Node()
            for k, v in val.items():
                setattr(node, k, self.value_to_node(v))
            return node
        return val

    def make_member(self, name):
        return self.value_to_node({
            "type": "MemberExpression",
            "object": {
                "name": "$bm_transform",
                "type": "Identifier"
            },
            "property": {
                "name": name,
                "type": "Identifier"
            }
        })

    def assign_variable(self, right, expr=True):
        data = {
            "left": {
                "name": "$bm_rt",
                "type": "Identifier"
            },
            "type": "AssignmentExpression",
            "operator": "=",
            "right": right,
        }
        if expr:
            data = {
                "type": "ExpressionStatement",
                "expression": data
            }
        return self.value_to_node(data)

    def declare(self, name):
        if name not in self.scope:
            self.scope.add(name)
            self.stack[-1].add(name)

    def push(self, vars):
        self.stack.append(set(vars))

    def pop(self):
        popped = self.stack.pop()
        for item in popped:
            self.scope.remove(item)

    def process(self, node):
        node = self.process_node(node, False)
        if self.undeclared:
            node.body.insert(0, self.value_to_node({
                "type": "VariableDeclaration",
                "declarations": [
                    {
                        "type": "VariableDeclarator",
                        "id": {
                            "type": "Identifier",
                            "name": name
                        }
                    }
                    for name in self.undeclared
                ],
                "kind": "var"
            }))

        return node

    def process_node(self, node, in_member):
        scopes = self.on_node_enter(node)

        member = node.type == "MemberExpression"

        if scopes is not None:
            for name, value in node.items():
                is_scope = name in scopes
                if is_scope:
                    self.push(scopes[name])

                if isinstance(value, esprima.nodes.Node):
                    setattr(node, name, self.process_node(value, member))
                elif isinstance(value, list):
                    setattr(node, name, [self.process_node(v, member) for v in value])

                if is_scope:
                    self.pop()

        node = self.on_node_leave(node, in_member)
        return node

    def gather_params(self, params, names):
        for n in params:
            if n.type == "Identifier":
                names.append(n.name)
            elif n.type == "AssignmentPattern":
                self.gather_params([n.left], names)
            elif n.type == "RestElement":
                self.gather_params([n.argument], names)

    def on_node_enter(self, node):
        if node.type == "VariableDeclarator":
            self.declare(node.id.name)
        elif node.type == "BlockStatement":
            return {"body": []}
        elif node.type == "ArrowFunctionExpression" or node.type == "FunctionDeclaration":
            params = self.gather_params(node.params, [])
            return {"params": params, "body": params}
        elif node.type == "MemberExpression":
            if node.property.type == "Identifier":
                if node.property.name == "name":
                    node.property.name = "_name"
            elif node.property.type == "Literal":
                if node.property.value == "name":
                    node.property.value = "_name"
                    node.property.raw = "'_name'"

        return {}

    def on_node_leave(self, node, member):
        if node.type == "Identifier" and not member:
            if node.name in self.reserved_properties and node.name not in self.scope:
                return self.make_member(node.name)
        elif node.type == "ExpressionStatement":
            if node.expression.type == 'SequenceExpression':
                ass = self.assign_variable(node.expression.expressions[-1], False)
                node.expression.expressions[-1] = ass
            else:
                return self.assign_variable(node.expression)
        elif node.type == "BinaryExpression":
            rep = self.op_to_func.get(node.operator)
            if rep:
                return self.value_to_node({
                    "type": "CallExpression",
                    "arguments": [node.left, node.right],
                    "callee": {
                        "name": rep,
                        "type": "Identifier"
                    }
                })
        elif node.type == "AssignmentExpression":
            if node.left and node.left.name and node.left.name != "value" and node.left.name not in self.scope:
                self.undeclared.add(node.left.name)

        return node


def process_expression(string):
    """!
    @brief converts expressions the same was as the bodymovin plugin
    """
    parsed = esprima.parseScript(string)
    Converter().process(parsed)
    return "var $bm_rt;\n" + escodegen.generate(parsed)



import io
import json
import enum
import base64
from xml.etree import ElementTree

from PIL import ImageCms

from ...nvector import NVector

from .riff import RiffParser, StructuredReader, BigEndian, StructuredData, sint
from .gradient_xml import xml_value_to_python
from .cos import CosParser


class ListType(enum.Enum):
    Shape = enum.auto()
    Keyframe = enum.auto()
    Other = enum.auto()


class KeyframeType(enum.Enum):
    MultiDimensional = enum.auto()
    Position = enum.auto()
    NoValue = enum.auto()
    Color = enum.auto()


class EssentialType(enum.Enum):
    Scalar = 2
    Color = 4
    Position = 5
    Comment = 8
    MultiDimensional = 9
    Group = 10
    Enum = 13


class AepParser(RiffParser):
    utf8_containers = ["tdsn", "fnam", "pdnm"]

    def __init__(self, file):
        if file is not None:
            super().__init__(file)

            if self.header.format != "Egg!":
                raise Exception("Not an AEP file")
        else:
            # XML initialization
            self.end = 0
            self.endian = BigEndian()

        self.chunk_parsers = {
            "Utf8": AepParser.read_utf8,
            "alas": AepParser.read_utf8,
            "tdmn": AepParser.read_mn,
            "cdta": AepParser.read_cdta,
            "ldta": AepParser.read_ldta,
            "idta": AepParser.read_idta,
            "tdb4": AepParser.read_tdb4,
            "cdat": AepParser.read_cdat,
            "lhd3": AepParser.read_lhd3,
            "ldat": AepParser.read_ldat,
            "ppSn": RiffParser.read_float,
            "tdum": RiffParser.read_float,
            "tduM": RiffParser.read_float,
            "tdsb": AepParser.read_number,
            "pprf": AepParser.read_pprf,
            "fvdv": AepParser.read_number,
            "ftts": AepParser.read_number,
            "fifl": AepParser.read_number,
            "fipc": AepParser.read_number,
            "fiop": AepParser.read_number,
            "foac": AepParser.read_number,
            "fiac": AepParser.read_number,
            "wsnm": AepParser.read_utf16,
            "fcid": AepParser.read_number,
            "fovc": AepParser.read_number,
            "fovi": AepParser.read_number,
            "fits": AepParser.read_number,
            "fivc": AepParser.read_number,
            "fivi": AepParser.read_number,
            "fidi": AepParser.read_number,
            "fimr": AepParser.read_number,
            "CsCt": AepParser.read_number,
            "CapL": AepParser.read_number,
            "CcCt": AepParser.read_number,
            "CprC": AepParser.read_number,
            "mrid": AepParser.read_number,
            "numS": AepParser.read_number,
            "shph": AepParser.read_shph,
            "otda": AepParser.read_otda,
            "opti": AepParser.read_opti,
            "sspc": AepParser.read_sspc,
            "parn": AepParser.read_number,
            "pard": AepParser.read_pard,
            "btdk": AepParser.read_btdk,
            "NmHd": AepParser.read_nmhd,
            "tdpi": AepParser.read_number,
            "tdps": lambda self, len: self.endian.decode_2comp(self.file.read(len)),
            "tdli": AepParser.read_number,
            "cmta": AepParser.read_mn,
            "mkif": AepParser.read_mkif,
            "blsv": AepParser.read_number,
            "blsi": AepParser.read_number,
            "CCId": AepParser.read_number,
            "CLId": AepParser.read_number,
            "CDim": AepParser.read_number,
            "CTyp": AepParser.read_ctyp,
            "StVS": AepParser.read_stvs,
            "Smin": AepParser.read_essential_value,
            "Smax": AepParser.read_essential_value,
            "CVal": AepParser.read_essential_value,
            "CDef": AepParser.read_essential_value,
            "fips": AepParser.read_fips,
            "svap": AepParser.read_svap,
            "head": AepParser.read_head,
        }
        for ch in self.utf8_containers:
            self.chunk_parsers[ch] = RiffParser.read_sub_chunks
        self.weird_lists = {
            "btdk": AepParser.read_btdk,
        }
        self.prop_dimension = None
        self.list_type = ListType.Other
        self.keyframe_type = KeyframeType.MultiDimensional
        self.ldat_size = 0
        self.keep_ldat_bytes = False
        self.essential_type = EssentialType.MultiDimensional

    def read_shph(self, length):
        reader = StructuredReader(self, length)
        reader.skip(3)
        reader.read_attribute("attrs", 1, bytes)
        # Relative to the layer position
        reader.read_attribute_array("top_left", 2, 4, float)
        reader.read_attribute_array("bottom_right", 2, 4, float)
        reader.skip(4)
        reader.finalize()
        reader.attr_bit("open", 0, 3)
        return reader.value

    def read_mn(self, length):
        return self.read(length).strip(b"\0").decode("utf8")

    def read_pprf(self, length):
        return ImageCms.ImageCmsProfile(io.BytesIO(self.read(length)))

    def read_tdb4(self, length):
        reader = StructuredReader(self, length)
        reader.skip(2) # db 99
        reader.read_attribute("components", 2, int)
        reader.read_attribute("attrs", 2, bytes)
        reader.read_attribute("", 1, bytes) # 00
        reader.read_attribute("", 1, bytes) # 03 iff position, else 00
        reader.read_attribute("", 2, bytes) # ffff 0002 0001
        reader.read_attribute("", 2, bytes) # ffff 0004 0000 0007
        reader.read_attribute("time_den", 4, int)
        reader.read_attribute("", 8, float) # most of the time 0.0001
        reader.read_attribute("", 8, float) # most of the time 1.0, sometimes 1.777
        reader.read_attribute("", 8, float) # 1.0
        reader.read_attribute("", 8, float) # 1.0
        reader.read_attribute("", 8, float) # 1.0
        reader.read_attribute("type", 4, bytes)
        reader.read_attribute("", 1, bytes) # Seems somehow correlated with the previous byte
        reader.read_attribute("", 7, bytes) # bunch of 00
        reader.read_attribute("animated", 1, int) # 01 iff animated
        reader.read_attribute("", 7, bytes) # bunch of 00
        reader.read_attribute("", 4, bytes) # Usually 0, probs flags
        reader.read_attribute("", 4, int) # most likely flags, only last byte seems to contain data
        reader.read_attribute("", 8, float) # always 0.0
        reader.read_attribute("", 8, float) # mostly 0.0, sometimes 0.333
        reader.read_attribute("", 8, float) # always 0.0
        reader.read_attribute("", 8, float) # mostly 0.0, sometimes 0.333
        reader.read_attribute("", 4, bytes) # probs some flags
        reader.read_attribute("", 4, bytes) # probs some flags

        reader.finalize()
        reader.attr_bit("position", 1, 3)
        reader.attr_bit("static", 1, 0)

        reader.attr_bit("no_value", 1, 0, "type")
        reader.attr_bit("color", 3, 0, "type")
        reader.attr_bit("integer", 3, 2, "type")
        data = reader.value

        self.prop_dimension = data.components
        if data.position:
            self.keyframe_type = KeyframeType.Position
        elif data.color:
            self.keyframe_type = KeyframeType.Color
        elif data.no_value:
            self.keyframe_type = KeyframeType.NoValue
        else:
            self.keyframe_type = KeyframeType.MultiDimensional

        return data

    def read_cdat(self, length):
        dim = self.prop_dimension
        self.prop_dimension = None

        if dim is None or length < dim * 8:
            return self.read(length)

        value = StructuredReader(self, length)
        value.read_attribute_array("value", dim, 8, float)
        if value.to_read % 8 == 0:
            value.read_attribute_array("", value.to_read // 8, 8, float)
        value.finalize()
        return value.value

    def read_cdta(self, length):
        reader = StructuredReader(self, length)

        reader.read_attribute("resolution_x", 2, int)
        reader.read_attribute("resolution_y", 2, int)
        reader.read_attribute("framerate_den", 4, int)
        reader.read_attribute("framerate_num", 4, int)
        reader.fraction("framerate")
        reader.skip(8)
        reader.read_attribute("playhead_num", 4, sint)
        reader.read_attribute("playhead_den", 4, int)
        reader.fraction("playhead")
        reader.read_attribute("start_time_num", 4, sint)
        reader.read_attribute("start_time_den", 4, int)
        reader.fraction("start_time")
        reader.read_attribute("end_time_num", 4, sint)
        reader.read_attribute("end_time_den", 4, int)
        reader.fraction("end_time")
        reader.read_attribute("comp_duration_num", 4, sint)
        reader.read_attribute("comp_duration_den", 4, int)
        reader.fraction("comp_duration")
        reader.read_attribute_array("color", 3, 1, int)
        reader.skip(84)

        reader.read_attribute("attrs", 1, bytes)
        reader.attr_bit("shy", 0, 0)
        reader.attr_bit("motion_blur", 0, 3)
        reader.attr_bit("frame_blending", 0, 4)
        reader.attr_bit("preserve_framerate", 0, 5)
        reader.attr_bit("preserve_resolution", 0, 7)

        reader.read_attribute("width", 2, int)
        reader.read_attribute("height", 2, int)
        reader.read_attribute("pixel_ratio_width", 4, int)
        reader.read_attribute("pixel_ratio_height", 4, int)
        reader.skip(4)
        reader.read_attribute("frame_rate_int", 2, int)
        reader.skip(16)
        reader.read_attribute("shutter_angle", 2, int)
        reader.read_attribute("shutter_phase", 4, sint)
        reader.skip(16)
        reader.read_attribute("samples_limit", 4, int)
        reader.read_attribute("samples_per_frame", 4, int)

        reader.finalize()
        return reader.value

    def read_ldta(self, length):
        reader = StructuredReader(self, length)

        # 0
        reader.read_attribute("layer_id", 4, int)
        # 4
        reader.read_attribute("quality", 2, int)
        reader.skip(2)
        # 8
        reader.read_attribute("stretch_num", 4, int)
        # 12
        reader.read_attribute("start_time_num", 4, sint)
        reader.read_attribute("start_time_den", 4, int)
        # 20
        reader.read_attribute("in_time_num", 4, sint)
        reader.read_attribute("in_time_den", 4, int)
        # 28
        reader.read_attribute("out_time_num", 4, sint)
        reader.read_attribute("out_time_den", 4, int)
        # 37
        reader.read_attribute("attrs", 4, bytes)
        # 40
        reader.read_attribute("source_id", 4, int)
        # 44
        reader.skip(17)
        reader.read_attribute("color", 1, int)
        reader.skip(2)
        # 64
        reader.read_attribute_string0("name", 32)
        # 96
        reader.skip(11)
        reader.read_attribute("matte_mode", 1, int)
        reader.read_attribute("stretch_den", 4, int)
        reader.skip(19)
        # 131
        reader.read_attribute("type", 1, int)
        # 132
        reader.read_attribute("parent_id", 4, int)
        reader.skip(24)
        if reader.to_read > 4:
            reader.read_attribute("track_matte_id", 4, int)

        reader.finalize()

        reader.attr_bit("bicubic", 1, 6)
        reader.attr_bit("guide", 2, 1)

        reader.attr_bit("auto_orient", 2, 0)
        reader.attr_bit("adjustment", 2, 1)
        reader.attr_bit("ddd", 2, 2)
        reader.attr_bit("solo", 2, 3)
        reader.attr_bit("null", 2, 7)

        reader.attr_bit("visible", 3, 0)
        reader.attr_bit("effects", 3, 2)
        reader.attr_bit("motion_blur", 3, 3)
        reader.attr_bit("locked", 3, 5)
        reader.attr_bit("shy", 3, 6)
        reader.attr_bit("rasterize", 3, 7)

        reader.fraction("stretch")
        reader.fraction("start_time")
        reader.fraction("in_time")
        reader.fraction("out_time")

        return reader.value

    def read_idta(self, length):
        reader = StructuredReader(self, length)

        reader.read_attribute("type", 2, int)
        reader.skip(14)
        reader.read_attribute("id", 4, int)
        reader.skip(38)
        reader.read_attribute("color", 1, int)

        reader.finalize()

        reader.value.type_name = "?"
        if reader.value.type == 4:
            reader.value.type_name = "composition"
        elif reader.value.type == 1:
            reader.value.type_name = "folder"
        elif reader.value.type == 7:
            reader.value.type_name = "footage"

        return reader.value

    def read_lhd3(self, length):
        reader = StructuredReader(self, length)
        reader.skip(10)
        reader.read_attribute("count", 2, int)
        reader.skip(6)
        reader.read_attribute("item_size", 2, int)
        reader.skip(3)
        reader.read_attribute("type", 1, int)
        reader.finalize()
        self.ldat_size = reader.value.count
        return reader.value

    def read_ldat_item_bezier(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("x", 4, float)
        reader.read_attribute("y", 4, float)
        reader.finalize()
        return reader.value

    def read_ldat_keyframe_common(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("time_num", 4, sint)
        reader.skip(1)
        reader.read_attribute("ease_mode", 1, int)
        reader.read_attribute("color", 1, int)
        reader.read_attribute("attrs", 1, bytes)

        reader.value.linear = reader.value.ease_mode == 1
        reader.value.hold = reader.value.ease_mode == 2
        reader.value.ease = reader.value.ease_mode == 3

        reader.attr_bit("continuous_bezier", 0, 3)
        reader.attr_bit("auto_bezier", 0, 4)
        reader.attr_bit("roving", 0, 5)

        return reader

    def read_ldat_keyframe_position(self, length):
        reader = self.read_ldat_keyframe_common(length)
        reader.skip(8)
        reader.read_attribute("", 8, float)
        reader.read_attribute("in_speed", 8, float)
        reader.read_attribute("in_influence", 8, float)
        reader.read_attribute("out_speed", 8, float)
        reader.read_attribute("out_influence", 8, float)
        reader.read_attribute_array("value", self.prop_dimension, 8, float)
        reader.read_attribute_array("pos_tan_in", self.prop_dimension, 8, float)
        reader.read_attribute_array("pos_tan_out", self.prop_dimension, 8, float)
        reader.finalize()
        return reader.value

    def read_ldat_keyframe_no_value(self, length):
        reader = self.read_ldat_keyframe_common(length)
        reader.skip(8)
        if reader.to_read >= 8 * 6:
            reader.read_attribute("", 8, float)
            reader.read_attribute("in_speed", 8, float)
            reader.read_attribute("in_influence", 8, float)
            reader.read_attribute("out_speed", 8, float)
            reader.read_attribute("out_influence", 8, float)
            reader.skip(8)
        reader.finalize()
        return reader.value

    def read_ldat_keyframe_multidimensional(self, length):
        reader = self.read_ldat_keyframe_common(length)
        reader.read_attribute_array("value", self.prop_dimension, 8, float)
        reader.read_attribute_array("in_speed", self.prop_dimension, 8, float)
        reader.read_attribute_array("in_influence", self.prop_dimension, 8, float)
        reader.read_attribute_array("out_speed", self.prop_dimension, 8, float)
        reader.read_attribute_array("out_influence", self.prop_dimension, 8, float)
        reader.finalize()
        return reader.value

    def read_ldat_keyframe_color(self, length):
        reader = self.read_ldat_keyframe_common(length)
        reader.read_attribute_array("", 2, 8, float)
        reader.read_attribute("in_speed", 8, float)
        reader.read_attribute("in_influence", 8, float)
        reader.read_attribute("out_speed", 8, float)
        reader.read_attribute("out_influence", 8, float)
        reader.read_attribute_array("value", self.prop_dimension, 8, float)
        reader.read_attribute_array("", reader.to_read // 8, 8, float)
        reader.finalize()
        return reader.value

    def read_ldat_keyframe_unknown(self, length):
        reader = self.read_ldat_keyframe_common(length)
        reader.finalize()
        return reader.value

    def read_ldat_item_raw(self, length):
        return self.read(length)

    def read_ldat(self, length):
        item_func = None

        if self.list_type == ListType.Other:
            item_func = self.read_ldat_item_raw
            array_name = "items"
        elif self.list_type == ListType.Shape:
            item_func = self.read_ldat_item_bezier
            array_name = "points"
        elif self.list_type == ListType.Keyframe:
            array_name = "keyframes"

            if self.keyframe_type == KeyframeType.Position:
                item_func = self.read_ldat_keyframe_position
            elif self.keyframe_type == KeyframeType.NoValue:
                item_func = self.read_ldat_keyframe_no_value
            elif self.keyframe_type == KeyframeType.Color:
                item_func = self.read_ldat_keyframe_color
            else:
                item_func = self.read_ldat_keyframe_multidimensional

            #item_func = self.read_ldat_keyframe_unknown

            self.keyframe_type = KeyframeType.MultiDimensional

        item_count = self.ldat_size
        item_size = length // item_count
        leftover = length % item_count
        value = StructuredData()
        items = []

        for i in range(item_count):
            item = item_func(item_size)
            if self.keep_ldat_bytes:
                if isinstance(item, bytes):
                    value.raw_bytes += item
                else:
                    value.raw_bytes += item.raw_bytes
            items.append(item)

        setattr(value, array_name, items)

        if leftover:
            value._leftover = self.read(leftover)

        return value

    def on_list_start(self, type):
        if type == "shap":
            self.list_type = ListType.Shape
        elif type == "tdbs":
            self.list_type = ListType.Keyframe

    def on_list_end(self, type):
        if type == "shap":
            self.list_type = ListType.Other
        elif type == "tdbs":
            self.list_type = ListType.Other

    def read_utf8(self, length):
        data = self.read(length).decode("utf8")
        if data.startswith("<?xml version='1.0'?>"):
            dom = ElementTree.fromstring(data)
            if dom.tag == "prop.map":
                return xml_value_to_python(dom)
            else:
                return dom
        elif data.startswith("{") and data.endswith("}"):
            try:
                jdata = json.loads(data)
                if "baseColorProfile" in jdata:
                    jdata["baseColorProfile"]["colorProfileData"] = ImageCms.ImageCmsProfile(io.BytesIO(
                        base64.b64decode(jdata["baseColorProfile"]["colorProfileData"])
                    ))
                return jdata
            except Exception:
                pass

        return data

    def read_utf16(self, length):
        return self.read(length).decode("utf16")

    def read_otda(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("x", 8, float)
        reader.read_attribute("y", 8, float)
        reader.read_attribute("z", 8, float)
        reader.finalize()
        return reader.value

    def read_opti(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("type", 4, str)
        if reader.value.type == "Soli":
            reader.skip(6)
            reader.read_attribute("a", 4, float)
            reader.read_attribute("r", 4, float)
            reader.read_attribute("g", 4, float)
            reader.read_attribute("b", 4, float)
            reader.read_attribute_string0("name", 256)
        reader.finalize()
        return reader.value

    def read_sspc(self, length):
        reader = StructuredReader(self, length)
        reader.skip(32)
        reader.read_attribute("width", 2, int)
        reader.skip(2)
        reader.read_attribute("height", 2, int)
        reader.skip(2)
        reader.read_attribute("frame_count", 2, int)
        reader.finalize()
        return reader.value

    def read_pard(self, length):
        reader = StructuredReader(self, length)
        reader.skip(15)
        reader.read_attribute("type", 1, int)
        reader.read_attribute_string0("name", 32)
        reader.skip(8)

        if reader.value.type == 2 or reader.value.type == 3: # Scalar / Angle
            reader.read_attribute("last_value", 4, sint)
            reader.value.last_value = reader.value.last_value / 0x10000
            reader.skip(4)
            reader.skip(64)
            reader.skip(4)
            reader.read_attribute("min_value", 2, sint)
            reader.skip(2)
            reader.read_attribute("max_value", 2, sint)
        elif reader.value.type == 4: # Boolean
            reader.read_attribute("last_value", 4, int)
            reader.read_attribute("default_value", 1, int)
            reader.skip(3)
            reader.skip(44)
            reader.read_attribute("", 4, float)
            reader.skip(4)
            reader.read_attribute("", 4, float)
        elif reader.value.type == 5: # Color
            reader.read_attribute_array("last_value", 4, 1, int)
            reader.read_attribute_array("default_value", 4, 1, int)
            reader.skip(64)
            reader.read_attribute_array("max_value", 4, 1, int)
        elif reader.value.type == 6: # 2D
            reader.read_attribute("last_value_x", 4, sint)
            reader.read_attribute("last_value_y", 4, sint)
            reader.value.last_value = NVector(reader.value.last_value_x, reader.value.last_value_y) / 0x80
        elif reader.value.type == 7: # Enum
            reader.read_attribute("last_value", 4, int)
            reader.read_attribute("option_count", 2, int)
            reader.read_attribute("default_value", 2, int)
            reader.skip(44)
            reader.read_attribute("", 4, float)
            reader.skip(4)
            reader.read_attribute("", 4, float)
        elif reader.value.type == 10: # Slider
            reader.read_attribute("last_value", 8, float)
            reader.skip(44)
            reader.read_attribute("", 4, float)
            reader.skip(4)
            reader.read_attribute("max_value", 4, float)
        elif reader.value.type == 18: # 3D Point
            reader.read_attribute("last_value_x", 8, float)
            reader.read_attribute("last_value_y", 8, float)
            reader.read_attribute("last_value_z", 8, float)
            reader.value.last_value = NVector(
                reader.value.last_value_x,
                reader.value.last_value_y,
                reader.value.last_value_z
            ) * 512
            reader.read_attribute("", 8, float)
            reader.read_attribute("", 8, float)

        reader.finalize()
        return reader.value

    def read_btdk(self, length):
        parser = CosParser(self.file, length)
        return parser.parse()

    def read_nmhd(self, length):
        reader = StructuredReader(self, length)
        reader.skip(3)
        reader.read_attribute("attrs", 1, bytes)
        reader.skip(4)
        reader.read_attribute("duration_num", 4, sint)
        reader.read_attribute("duration_den", 4, int)
        reader.fraction("duration")
        reader.read_attribute("color", 1, int)
        reader.finalize()

        reader.attr_bit("protected", 0, 1)
        return reader.value

    def read_mkif(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("inverted", 1, int)
        reader.read_attribute("locked", 1, int)
        reader.skip(4)
        reader.read_attribute("mode", 2, int)
        reader.finalize()
        return reader.value

    def read_ctyp(self, length):
        data = self.read_number(length)
        self.essential_type = EssentialType(data)
        return data

    def read_stvs(self, length):
        reader = StructuredReader(self, length)
        reader.read_attribute("count", 1, int)
        reader.finalize()
        return reader.value

    def read_essential_value(self, length):
        reader = StructuredReader(self, length)

        if self.essential_type == EssentialType.Scalar:
            reader.read_attribute("value", 8, float)
        elif self.essential_type == EssentialType.Color:
            reader.read_attribute_array("value", 4, 4, float)
        elif self.essential_type == EssentialType.Position:
            reader.read_attribute_array("value", 2, 8, float)
        elif self.essential_type == EssentialType.MultiDimensional:
            reader.read_attribute_array("value", 4, 8, float)
        elif self.essential_type == EssentialType.Enum:
            reader.read_attribute("value", 4, int)
        else:
            return self.read(length)

        reader.finalize()
        return reader.value.value

    def read_fips(self, length):
        reader = StructuredReader(self, length)
        reader.skip(12)
        reader.skip(3)
        reader.read_attribute("attrs", 1, bytes)
        reader.attr_bit("alpha_grid", 0, 7)
        reader.finalize()
        return reader.value

    def read_ae_version(self, reader: StructuredReader):
        reader.read_attribute("version_bits", 4, bytes)
        # See https://github.com/tinogithub/aftereffects-version-check/blob/main/images/byte-scheme.PNG
        value = self.endian.decode(reader.value.version_bits)
        # bits 0-7 build
        reader.value.version_build = value & 0xff
        value >>= 8
        # bit 8 ?
        value >>= 1
        # bit 9 beta
        reader.value.version_beta = value & 1
        value >>= 1
        # bit 10 ?
        value >>= 1
        # bit 11-14 patch
        reader.value.version_patch = value & 0xf
        value >>= 4
        # bit 15-18 minor
        reader.value.version_minor = value & 0xf
        value >>= 4
        # bit 19-21 major low bits
        major_low = value & 0x7
        value >>= 3
        # bit 22-25 os
        reader.value.version_os = value & 0xf
        value >>= 4
        if reader.value.version_os == 0b1110:
            os_string = "Mac Arm 64"
        elif reader.value.version_os == 0b1100:
            os_string = "Win"
        elif reader.value.version_os == 0b1101:
            os_string = "Mac"
        else:
            os_string = "Unknown"
        reader.value.version_os_name = os_string
        # bit 26-30 major high bits
        major_high = (value & 0x1f) << 3
        value >>= 6
        reader.value.version_major = major_high | major_low
        reader.value.version = "%s.%s.%s" % (
            reader.value.version_major,
            reader.value.version_minor,
            reader.value.version_patch,
        )
        reader.value.version_full = "v%s%s (%s) (Build %s)" % (
            reader.value.version,
            " BETA" if reader.value.version_beta else "",
            os_string,
            reader.value.version_build,
        )

    def read_svap(self, length):
        reader = StructuredReader(self, length)
        self.read_ae_version(reader)
        reader.finalize()
        return reader.value

    def read_head(self, length):
        reader = StructuredReader(self, length)
        reader.skip(4)
        self.read_ae_version(reader)
        reader.finalize()
        return reader.value



from ...nvector import NVector


def xml_value_to_python(element):
    if element.tag == "prop.map":
        return xml_value_to_python(element[0])
    elif element.tag == "prop.list":
        return xml_list_to_dict(element)
    elif element.tag == "array":
        return xml_array_to_list(element)
    elif element.tag == "int":
        return int(element.text)
    elif element.tag == "float":
        return float(element.text)
    elif element.tag == "string":
        return element.text
    else:
        return element


def xml_array_to_list(element):
    data = []
    for ch in element:
        if ch.tag != "array.type":
            data.append(xml_value_to_python(ch))
    return data


def xml_list_to_dict(element):
    data = {}
    for pair in element.findall("prop.pair"):
        key = None
        value = None
        for ch in pair:
            if ch.tag == "key":
                key = ch.text
            else:
                value = xml_value_to_python(ch)
        data[key] = value

    return data


def parse_gradient_xml(gradient, colors_prop):
    flat = []

    data = gradient["Gradient Color Data"]

    previous = None
    count = 0
    for stop in data["Color Stops"]["Stops List"].values():
        colors = stop["Stops Color"]
        if previous is not None:
            offset = previous[0] * (1 - previous[1]) + colors[0] * previous[1]
            count += 1
            flat += [offset, (previous[2] + colors[2]) / 2, (previous[3] + colors[3]) / 2, (previous[4] + colors[4]) / 2]

        count += 1
        flat += [colors[0], colors[2], colors[3], colors[4]]
        previous = colors

    previous = None
    for stop in data["Alpha Stops"]["Stops List"].values():
        alpha = stop["Stops Alpha"]
        if previous is not None:
            offset = previous[0] * (1 - previous[1]) + colors[0] * previous[1]
            flat += [offset, (previous[2] + colors[2])]
        flat += [alpha[0], alpha[2]]

    colors_prop.count = count

    return NVector(*flat)



import os
import subprocess
from . import objects, parsers, utils, exporters, nvector, importers
from .nvector import *
from .utils.color import Color

try:
    from .version import __version__
except ImportError:
    here = os.path.dirname(os.path.abspath(__file__))
    __version__ = "unknown"
    try:
        pipe = subprocess.Popen(
            ['git', 'describe', '--abbrev=0', '--tags'],
            cwd=here,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.PIPE
        )
    except FileNotFoundError:
        pass
    else:
        out, err = pipe.communicate()
        if pipe.returncode == 0:
            __version__ = out.strip()[1:].decode("ascii") + "+git"
        else:
            vfn = os.path.join(os.path.dirname(os.path.dirname(here)), "version")
            if os.path.exists(vfn):
                with open(vfn) as vf:
                    __version__ = vf.read().strip() + "+src"


try:
    version_tuple = tuple(map(int, __version__.split("+")[0].split("."))) if __version__ != "unknown" else (0, 0, 0)
except ValueError:
    version_tuple = (0, 0, 0)
    __version__ = "unknown"


__all__ = ["objects", "parsers", "utils", "exporters", "nvector", "NVector", "Point", "Color", "importers"]
